/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright © 2003 Atheros Communications, Inc.,  All Rights Reserved.
 */

/*
 * Initialization for ar531x SOC.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/serial.h>
#include <linux/types.h>
#include <linux/string.h>

#include <asm/reboot.h>
#include <asm/io.h>
#include <asm/time.h>
#include <asm/pgtable.h>
#include <asm/processor.h>
#include <asm/reboot.h>
#include <asm/system.h>
#include <asm/serial.h>

#include "ar531xlnx.h"

void
ar531x_restart(char *command)
{
    for(;;) {
        sysRegWrite(AR531X_RESET, AR531X_RESET_SYSTEM);
    }
}

void
ar531x_halt(void)
{
        printk(KERN_NOTICE "\n** You can safely turn off the power\n");
        while (1);
}

void
ar531x_power_off(void)
{
        ar531x_halt();
}

const char *
get_system_type(void)
{
	return "Atheros AR531X";
}

/*
 * This table is indexed by bits 5..4 of the CLOCKCTL1 register
 * to determine the predevisor value.
 */
static int CLOCKCTL1_PREDIVIDE_TABLE[4] = {
    1,
    2,
    4,
    5
};

unsigned int
ar531x_cpu_frequency(void)
{
	static unsigned int ar531x_calculated_cpu_freq;
        unsigned int clockctl1_predivide_mask;
        unsigned int clockctl1_predivide_shift;
        unsigned int clockctl1_multiplier_mask;
        unsigned int clockctl1_multiplier_shift;
        unsigned int clockctl1_doubler_mask;
        int wisoc_revision;

        /*
         * Trust the bootrom's idea of cpu frequency.
         */
        ar531x_calculated_cpu_freq = sysRegRead(AR5312_SCRATCH);
        if (ar531x_calculated_cpu_freq)
	    return ar531x_calculated_cpu_freq;

        wisoc_revision = (sysRegRead(AR531X_REV) & AR531X_REV_MAJ) >> AR531X_REV_MAJ_S;
        if (wisoc_revision == AR531X_REV_MAJ_AR2313) {
            clockctl1_predivide_mask = AR2313_CLOCKCTL1_PREDIVIDE_MASK;
            clockctl1_predivide_shift = AR2313_CLOCKCTL1_PREDIVIDE_SHIFT;
            clockctl1_multiplier_mask = AR2313_CLOCKCTL1_MULTIPLIER_MASK;
            clockctl1_multiplier_shift = AR2313_CLOCKCTL1_MULTIPLIER_SHIFT;
            clockctl1_doubler_mask = AR2313_CLOCKCTL1_DOUBLER_MASK;
        } else { /* AR5312 and AR2312 */
            clockctl1_predivide_mask = AR5312_CLOCKCTL1_PREDIVIDE_MASK;
            clockctl1_predivide_shift = AR5312_CLOCKCTL1_PREDIVIDE_SHIFT;
            clockctl1_multiplier_mask = AR5312_CLOCKCTL1_MULTIPLIER_MASK;
            clockctl1_multiplier_shift = AR5312_CLOCKCTL1_MULTIPLIER_SHIFT;
            clockctl1_doubler_mask = AR5312_CLOCKCTL1_DOUBLER_MASK;
        }

        /*
         * Clocking is derived from a fixed 40MHz input clock.
         *  cpuFreq = InputClock * MULT (where MULT is PLL multiplier)
         *
         *  sysFreq = cpuFreq / 4       (used for APB clock, serial,
         *                               flash, Timer, Watchdog Timer)
         *
         *  cntFreq = cpuFreq / 2       (use for CPU count/compare)
         *
         * So, for example, with a PLL multiplier of 5, we have
         *  cpuFrez = 200MHz
         *  sysFreq = 50MHz
         *  cntFreq = 100MHz
         *
         * We compute the CPU frequency, based on PLL settings.
         */
	if (ar531x_calculated_cpu_freq == 0) {
            unsigned int clockCtl1 = sysRegRead(AR5312_CLOCKCTL1);

            int preDivideSelect = (clockCtl1 & clockctl1_predivide_mask) >>
                                   clockctl1_predivide_shift;

            int preDivisor = CLOCKCTL1_PREDIVIDE_TABLE[preDivideSelect];

            int multiplier = (clockCtl1 & clockctl1_multiplier_mask) >>
                              clockctl1_multiplier_shift;

            if (clockCtl1 & clockctl1_doubler_mask) {
                multiplier = multiplier << 1;
            }

            ar531x_calculated_cpu_freq = (40000000 / preDivisor) * multiplier;
        }

	return ar531x_calculated_cpu_freq;
}

unsigned int
ar531x_sys_frequency(void)
{
	static unsigned int ar531x_calculated_sys_freq = 0;

	if (ar531x_calculated_sys_freq == 0) {
		ar531x_calculated_sys_freq = ar531x_cpu_frequency() / 4;
	}

	return ar531x_calculated_sys_freq;
}

struct ar531x_boarddata *ar531x_board_configuration = NULL;
char *ar531x_radio_configuration = NULL;

static void __init
flash_setup(void)
{
    UINT32 flash_ctl;
    int ar531x_flash_data_found;
    char *bd_config;

    /* Configure flash bank 0 */
    flash_ctl = FLASHCTL_E |
                FLASHCTL_AC_8M |
                FLASHCTL_RBLE |
                (0x01 << FLASHCTL_IDCY_S) |
                (0x07 << FLASHCTL_WST1_S) |
                (0x07 << FLASHCTL_WST2_S) |
                (sysRegRead(AR531X_FLASHCTL0) & FLASHCTL_MW);

    sysRegWrite(AR531X_FLASHCTL0, flash_ctl);

    /* Disable other flash banks */
    sysRegWrite(AR531X_FLASHCTL1,
                sysRegRead(AR531X_FLASHCTL1) & ~(FLASHCTL_E | FLASHCTL_AC));

    sysRegWrite(AR531X_FLASHCTL2,
                sysRegRead(AR531X_FLASHCTL2) & ~(FLASHCTL_E | FLASHCTL_AC));

#if 0
    /*
     * Find start of Board Configuration data, using heuristics:
     * Search back from the (aliased) end of flash by 0x1000 bytes
     * at a time until we find the string "5311", which marks the
     * start of Board Configuration.  Give up if we've searched
     * more than 500KB.
     */
    ar531x_flash_data_found = 0;
    for (bd_config = (char *)0xbffff000;
         bd_config > (char *)0xbff80000;
         bd_config -= 0x1000)
    {
        if ( *(int *)bd_config == AR531X_BD_MAGIC) {
            ar531x_flash_data_found = 1;
            break;
        }
    }

    if (!ar531x_flash_data_found) {
        printk("Could not find Board Configuration Data\n");
        return;
    }

    ar531x_board_configuration = (struct ar531x_boarddata *)bd_config;

    /* 
     * Now find the start of Radio Configuration data, using heuristics:
     * Search forward from Board Configuration data by 0x1000 bytes
     * at a time until we find non-0xffffffff.
     */
    ar531x_flash_data_found = 0;
    for (ar531x_radio_configuration = bd_config + 0x1000;
         ar531x_radio_configuration < (char *)0xbffff000;
         ar531x_radio_configuration += 0x1000)
    {
        if (*(int *)ar531x_radio_configuration != 0xffffffff) {
            ar531x_flash_data_found = 1;
            break;
        }
    }

    if (!ar531x_flash_data_found) {
        printk("Could not find Radio Configuration data\n");
    }
#endif
}

static char invalid_enet_MAC_addr[] = {0, 0, 0, 0, 0, 0};

/*
 * Fetch a pointer to an ethernet's MAC address
 * in the Board Configuration data (in flash).
 */
char *
enet_mac_address_get(int MACUnit)
{
    if (!ar531x_board_configuration)
        return invalid_enet_MAC_addr;

    if (MACUnit == 0) {
        return ar531x_board_configuration->enet0Mac;
    }

    if (MACUnit == 1) {
        return ar531x_board_configuration->enet1Mac;
    }

    printk("Invalid ethernet MAC unit number (%d)!\n", MACUnit);
    return invalid_enet_MAC_addr;
}


void __init
serial_setup(void)
{
	struct serial_struct s;

	memset(&s, 0, sizeof(s));

	s.flags = STD_COM_FLAGS;
	s.io_type = SERIAL_IO_MEM;
#if 0

	/* Calculate base baud clock from CPU frequency */

	s.baud_base = (ar531x_cpu_frequency() / 4) / 16;

#else

	s.baud_base = ar531x_sys_frequency()/16;

#endif

	s.irq = AR531X_MISC_IRQ_UART0;
	s.iomem_reg_shift = 2;
	s.iomem_base = (u8 *)0xbc000003;

	if (early_serial_setup(&s) != 0)
		printk(KERN_ERR "early_serial_setup failed\n");
}

extern int setup_irq(unsigned int irq, struct irqaction *irqaction);
static void __init
ar531x_timer_setup(struct irqaction *irq)
{
        unsigned int count;

	/* Usually irq is timer_irqaction (timer_interrupt) */
      	setup_irq(AR531X_IRQ_CPU_CLOCK, irq);

        /* to generate the first CPU timer interrupt */
        count = read_c0_count();
        write_c0_compare(count + 1000);
}

extern void (*board_time_init)(void);

static void __init
ar531x_time_init(void)
{
	mips_hpt_frequency = ar531x_cpu_frequency() / 2;
}

void __init
ar531x_setup(void)
{
	/* Clear any lingering AHB errors */
	sysRegRead(AR531X_PROCADDR);
	sysRegRead(AR531X_DMAADDR);

#ifdef CONFIG_AR531X_WATCHDOG
        (void)watchdog_stop();
#endif

        /* Disable data watchpoints */
        write_c0_watchlo0(0);

	board_time_init = ar531x_time_init;
        board_timer_setup = ar531x_timer_setup;

        _machine_restart = ar531x_restart;
        _machine_halt = ar531x_halt;
        _machine_power_off = ar531x_power_off;

        flash_setup();
        serial_setup();

        /* Select ENET1 over local bus */
        sysRegWrite(AR5312_LOCAL_BUS_CONFIG,
            	    sysRegRead(AR5312_LOCAL_BUS_CONFIG) &
            	        ~(AR5312_LBCONFIG_LOCAL_BUS));
}
