/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright © 2003 Atheros Communications, Inc.,  All Rights Reserved.
 */

/*
 * Support for the Watchdog Timer
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/interrupt.h>
#include <linux/irq.h>

#include "ar531xlnx.h"

/*
 * The AR5312/AR2313 hardware supports a hardware watchdog timer.
 * The timer is programmed to decrement until it hits zero.  Then
 * it generates either SYSTEM RESET or an NMI.  In order to avoid
 * one of these events, software must reset the watchdog timer
 * before it expires.
 *
 * Notes:
 *  -There's no locking on these functions.  watchdog_notify_alive can
 *   be called from anywhere with no need for locking, and the remaining
 *   functions are expected to be used only during startup.
 *  -When kgdb debugging is enabled, the watchdog timer is automatically
 *   disabled.
 */

static pid_t watchdog_pid;
static unsigned int ar531x_watchdog_timer_value = 0;

#define AR531X_WATCHDOG_MIN_MS 1000

/*
 * Macros to convert between "watchdog ticks" and milliseconds.
 * The watchdog counts down at the system frequency (1/4 CPU freq).
 */
#define wdticks_to_ms(wdticks) (((wdticks) * 1000) / ar531x_sys_frequency())
#define ms_to_wdticks(ms) (((ms) / 1000) * ar531x_sys_frequency())

static unsigned int min_timer_reached = 0xffffffffu;

/*
 * Notify the watchdog that we're alive.
 * Reset the watchdog timer.
 */
void
watchdog_notify_alive(void)
{
    unsigned int oldTimer;

    oldTimer = sysRegRead(AR531X_WD_TIMER);
    if (oldTimer < min_timer_reached)
        min_timer_reached = oldTimer;

    sysRegWrite(AR531X_WD_TIMER, ar531x_watchdog_timer_value);
    sysWbFlush();
}

/*
 * Returns the minimum value that the WD_TIMER ever reached.
 */
unsigned int
watchdog_min_timer_reached(void)
{
    return min_timer_reached;
}

/*
 * Returns
 *   0 if the watchdog is currently disabled
 *   1 if the watchdog is currently enabled
 */
int
watchdog_is_enabled(void)
{
    return (ar531x_watchdog_timer_value != 0);
}

/*
 * This is the thread the periodically resets the watchdog timer.
 * If this thread is unable to run for too long, the timer will
 * expire and the system will be interrupted.
 *
 * Note: In systems where constant large interrupt load is normal,
 * we may need to call watchdog_notify_alive from other places in
 * addition to -- or instead of -- from this thread.
 */
static int
ar531x_watchdog_thread(void *data)
{
    unsigned int milliseconds = (unsigned int)data;
    unsigned int sleeptime;

    /*
     * In order to avoid inappropriate watchdog expirations,
     * we'll be conservative and notify the watchdog that we're
     * alive at a rate that's 2 times the rate of expiration.
     */
    sleeptime = ((HZ*milliseconds >> 1)/1000);

    min_timer_reached = 0xffffffffu;

    for (;;) {
        watchdog_notify_alive();

        set_current_state(TASK_INTERRUPTIBLE);
        schedule_timeout(sleeptime);
    }
}

/*
 * Start the watchdog timer, or change the timer countdown value.
 */
int
watchdog_start(unsigned int milliseconds)
{
    if (kgdbEnabled())
        return -1;

    if (milliseconds < AR531X_WATCHDOG_MIN_MS)
        return -1;

    watchdog_stop(); /* just in case it's already running */

    ar531x_watchdog_timer_value = ms_to_wdticks(milliseconds);
    watchdog_notify_alive(); /* Initialize timer */

    /* Start a thread to periodically reset the watchdog timer */
    watchdog_pid = kernel_thread(ar531x_watchdog_thread, (void *)milliseconds, 0);
    if (watchdog_pid < 0) {
        ar531x_watchdog_timer_value = 0;
        return -1;
    }

    /*
     * Cause a system RESET on watchdog expiration.
     */
    sysRegWrite(AR531X_WD_CTRL, AR531X_WD_CTRL_RESET);

    return 0;
}


/*
 * Stop the watchdog timer.
 * Returns the old value of the timer, in milliseconds.
 */
int
watchdog_stop(void)
{
    sysRegWrite(AR531X_WD_CTRL, AR531X_WD_CTRL_IGNORE_EXPIRATION);

    if (watchdog_is_enabled()) {
        unsigned int OLD_watchdog_timer_value;

        OLD_watchdog_timer_value = ar531x_watchdog_timer_value;
        ar531x_watchdog_timer_value = 0;

        kill_proc(watchdog_pid, SIGINT, 1);
        return wdticks_to_ms(OLD_watchdog_timer_value);
    } else {
        return 0;
    }
}
