

/*** PORTIONS OF THIS SOURCE FILE ARE DERIVED FROM CODE COVERED UNDER THE
 * FOLLOWING LICENSE: */
/* Copyright (C) 1995-1998 Eric Young (eay@cryptsoft.com)
 * All rights reserved.
 *
 * This package is an SSL implementation written
 * by Eric Young (eay@cryptsoft.com).
 * The implementation was written so as to conform with Netscapes SSL.
 *
 * This library is free for commercial and non-commercial use as long as
 * the following conditions are aheared to.  The following conditions
 * apply to all code found in this distribution, be it the RC4, RSA,
 * lhash, DES, etc., code; not just the SSL code.  The SSL documentation
 * included with this distribution is covered by the same copyright terms
 * except that the holder is Tim Hudson (tjh@cryptsoft.com).
 *
 * Copyright remains Eric Young's, and as such any Copyright notices in
 * the code are not to be removed.
 * If this package is used in a product, Eric Young should be given attribution
 * as the author of the parts of the library used.
 * This can be in the form of a textual message at program startup or
 * in documentation (online or textual) provided with the package.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    "This product includes cryptographic software written by
 *     Eric Young (eay@cryptsoft.com)"
 *    The word 'cryptographic' can be left out if the rouines from the library
 *    being used are not cryptographic related :-).
 * 4. If you include any Windows specific code (or a derivative thereof) from
 *    the apps directory (application code) you must include an acknowledgement:
 *    "This product includes software written by Tim Hudson (tjh@cryptsoft.com)"
 *
 * THIS SOFTWARE IS PROVIDED BY ERIC YOUNG ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * The licence and distribution terms for any publically available version or
 * derivative of this code cannot be changed.  i.e. this code cannot simply be
 * copied and put under another distribution licence
 * [including the GNU Public Licence.]
 */

#include <jni.h>
#include "i802.h"

#include <stdio.h>
#include <openssl/ssl.h>
#include <openssl/tls1.h>
#include <openssl/hmac.h>
#include <openssl/md5.h>

#include <errno.h>

#include "jsyscall.h"
#include "jthread.h"

extern void i802_ssl_init(void);

#undef DEBUG

#ifdef DEBUG

#define I802_DBG_NONE		0x0000
#define I802_DBG_BIO		0x0001
#define I802_DBG_SETUP		0x0002
#define I802_DBG_DATA		0x0004
#define I802_DBG_CLOSE                                   0x0008

static int i802_debug_level = I802_DBG_NONE;

#define I802_DBG(level,args...) \
	do { \
		if (level & i802_debug_level) { \
			fprintf( stderr, ##args ); \
		} \
	} while (0)
#else
#define I802_DBG(...)
#endif

void throw_memory_exception(JNIEnv *env, char *msg);

/*
 * Timeout is global to all connections using this code since there is no
 * simple way to connect them at the moment.
 */
static int timeout = 15000;		/* default to 15 seconds */

/* Replacement BIO methods that use the Kaffe thread-safe I/O interface */

static int sock_free(BIO *a)
	{
	I802_DBG(I802_DBG_BIO, "sock free called (num = %d, init = %d shutdown = %d)\n",
	    a->num, a->init, a->shutdown);

	if (a == NULL)
		return(0);
	if (a->shutdown)
		{
		if (a->init)
			{
			I802_DBG(I802_DBG_CLOSE, "KSOCKCLOSE(%d)\n", a->num);
			KSOCKCLOSE(a->num);
			}
		a->init=0;
		a->flags=0;
		}
	return(1);
	}

static int sock_read(BIO *b, char *out, int outl)
	{
	int ret=0;
	int err;

	if (out != NULL)
		{
		errno = 0;
		jthread_seterrno(0);
		/* TODO: make the timeout connection specific */
		I802_DBG(I802_DBG_BIO, "KSOCKREAD called on %d timeout %d\n", b->num, timeout);
		err = KSOCKREAD(b->num,out,outl,timeout,&ret);
		I802_DBG(I802_DBG_BIO, "KSOCKREAD sd=%d returned %d error %d\n", b->num,  ret, err);
		BIO_clear_retry_flags(b);
		if (err !=0 )
			{
			jthread_seterrno(err);
			ret = -1;
			if (BIO_sock_should_retry(ret))
				BIO_set_retry_read(b);
			}
		else if (ret == 0)
			{
			/* Zero return implies 'end of file' so return an error */
			ret = -1;
			jthread_seterrno(EPIPE);
			}
		}
	return(ret);
	}

static int sock_write(BIO *b, const char *in, int inl)
	{
	int ret;
	int err;

	I802_DBG(I802_DBG_BIO, "sock write  (%d) called\n",b->num );

	errno = 0;
	jthread_seterrno(0);
	// Returns errno on error; zero on success
	err=KSOCKWRITE(b->num,in,inl,&ret);
	BIO_clear_retry_flags(b);
	if ( err != 0)
		{
		jthread_seterrno(err);
		ret = -1;
		if (BIO_sock_should_retry(ret))
			BIO_set_retry_write(b);
		}
	I802_DBG(I802_DBG_BIO, "exiting sock_write with return code: %d\n", ret );
	return(ret);
	}

static long sock_ctrl(BIO *b, int cmd, long num, void *ptr)
	{
	long ret=1;
	int *ip;

 	I802_DBG(I802_DBG_BIO, "sock ctrl called (b = %p cmd = %d num = %d ptr = %p)\n",
	    b, cmd, num, ptr);

	switch (cmd)
		{
	case BIO_CTRL_RESET:
		num=0;
	case BIO_C_FILE_SEEK:
		ret=0;
		break;
	case BIO_C_FILE_TELL:
	case BIO_CTRL_INFO:
		ret=0;
		break;
	case BIO_C_SET_FD:
		sock_free(b);
		b->num= *((int *)ptr);
		b->shutdown=(int)num;
		b->init=1;
		break;
	case BIO_C_GET_FD:
		if (b->init)
			{
			ip=(int *)ptr;
			if (ip != NULL) *ip=b->num;
			ret=b->num;
			}
		else
			ret= -1;
		break;
	case BIO_CTRL_GET_CLOSE:
		ret=b->shutdown;
		break;
	case BIO_CTRL_SET_CLOSE:
		b->shutdown=(int)num;
		break;
	case BIO_CTRL_PENDING:
	case BIO_CTRL_WPENDING:
		ret=0;
		break;
	case BIO_CTRL_DUP:
	case BIO_CTRL_FLUSH:
		ret=1;
		break;
	default:
		ret=0;
		break;
		}
	return(ret);
	}

static int sock_puts(BIO *bp, const char *str)
	{
	int n,ret;

	I802_DBG(I802_DBG_BIO, "sock puts called\n");

	n=strlen(str);
	ret=sock_write(bp,str,n);
	return(ret);
	}

/*
 * Override the standard OpenSSL BIO socket methods with the
 * Kaffe-safe versions
 */
static void i802_update_bio_method( void )
	{
	static int done = 0;

	if (!done)
		{
		BIO_METHOD *ptr = BIO_s_socket();
		ptr->bread = sock_read;
		ptr->bwrite = sock_write;
		ptr->bputs = sock_puts;
		ptr->ctrl = sock_ctrl;
		ptr->destroy = sock_free;
		done = 1;
		}
	}

/* Java Native Methods */

JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_createSSL(JNIEnv *env, jobject obj, jint ctx) {
		SSL_CTX *pctx;

		pctx = (SSL_CTX *)ctx;

		i802_update_bio_method();
		i802_ssl_init();

		return ((jint)SSL_new(pctx));
	}

JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_setFd(JNIEnv *env, jobject obj, jint ssl, jint socketd) {
		SSL *pssl;

		pssl = (SSL *) ssl;
		i802_ssl_init();

		I802_DBG(I802_DBG_SETUP, "Setting socket descriptor to %d\n", socketd);
		return (jint)(SSL_set_fd(pssl, socketd));
	}


/**
 * Accept connection on underlying socket.
 * @return 1 if accepted connection, 0 if need to try again, < 0 if error
 * */
JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_accept(JNIEnv *env, jobject obj, jint ssl) {
		SSL *pssl;
		int val, err;

		i802_ssl_init();
		pssl = (SSL *) ssl;
		I802_DBG(I802_DBG_SETUP, "Java_com_instant802_cldc_ssl_SSL_accept: calling SSL_accept\n");
		val = SSL_accept(pssl);
		I802_DBG(I802_DBG_SETUP, "Java_com_instant802_cldc_ssl_SSL_accept: SSL_accept returned %d\n", val);
		if (val != 1) {
			err = SSL_get_error(pssl, val);
			if ((err == SSL_ERROR_WANT_READ) || (err == SSL_ERROR_WANT_WRITE)) {
				/*
				 * If either of these happens, try again
				 */
				val = -err;
			} else if (err == SSL_ERROR_SYSCALL) {
				/*
				 * If it was a system call error, get the thread's errno
				 * and check for EAGAIN or EINTR that would indicate that
				 * we should try again.
				 */
				err = jthread_geterrno();
				if ((err == EAGAIN) || (err == EINTR)) {
					/* Timeout error */
					val = 0;
				} else {
					/* Must have been a real error */
					val = -1;
				}
			} else {
				/* Unknown error: report it as a fatal error */
				val = -1;
			}
		}
		I802_DBG(I802_DBG_SETUP, "Java_com_instant802_cldc_ssl_SSL_accept: returning %d\n", val);
		return (jint)val;
	}
/**
 * Connect. Returns 1 is success, 0 if need to try again. -1 if fatal error
 * -SSL_ERROR_WANT_READ or -SSL_ERROR_WANT_WRITE for the special OpenSSL cases
 */
JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_connect(JNIEnv *env, jobject obj, jint ssl) {
		SSL *pssl;
		int val, err;

		i802_ssl_init();
		pssl = (SSL *) ssl;
		// SSL_set_connect_state(pssl);
		I802_DBG(I802_DBG_SETUP, "Java_com_instant802_cldc_ssl_SSL_connect: calling SSL_connect\n");
		val = SSL_connect(pssl);
		I802_DBG(I802_DBG_SETUP, "Java_com_instant802_cldc_ssl_SSL_connect: SSL_connect returned %d\n", val);
		if (val != 1) {
			err = SSL_get_error(pssl, val);

			if ((err == SSL_ERROR_WANT_READ) || (err == SSL_ERROR_WANT_WRITE)) {
				/*
				 * If either of these happens, try again
				 */
				val = -err;
			} else if (err == SSL_ERROR_SYSCALL) {
				/*
				 * If it was a system call error, get the thread's errno
				 * and check for EAGAIN or EINTR that would indicate that
				 * we should try again.
				 */
				err = jthread_geterrno();
				if ((err == EAGAIN) || (err == EINTR)) {
					/* Timeout error */
					val = 0;
				} else {
					/* Must have been a real error */
					val = -1;
				}
			} else {
				/* Unknown error: report it as a fatal error */
				val = -1;
			}
		}
		I802_DBG(I802_DBG_SETUP, "Java_com_instant802_cldc_ssl_SSL_connect: returning %d\n", val);
		return (jint) val;
	}

/**
 * Non blocking read. Read through SSL. returns >0 num
 * of bytes read, 0 if try again, -1 if fatal error, 
 * -SSL_ERROR_WANT_READ or -SSL_ERROR_WANT_WRITE for the special OpenSSL cases

 */
JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_read(JNIEnv *env, jobject obj, jint ssl, jbyteArray jbuf, jint buflen, jint t) {
		SSL *pssl;
		int val, err;
		unsigned char *pbuf;

		pssl = (SSL *) ssl;
		i802_ssl_init();

		/* TODO: make the timeout connection specific */
		timeout = t;

		pbuf = (unsigned char *) malloc(buflen);
		if (NULL == pbuf) {
			throw_memory_exception(env, "SSL_read out of memory");
			return (jint) -1;
		}
		ERR_clear_error();
		I802_DBG(I802_DBG_DATA, "Java_com_instant802_cldc_ssl_SSL_read: calling SSL_read (timeout = %d)\n", timeout);
		val = SSL_read(pssl, pbuf, buflen);
		I802_DBG(I802_DBG_DATA, "Java_com_instant802_cldc_ssl_SSL_read: SSL_read returned %d\n", val);
		if (val > 0) {
			(*env)->SetByteArrayRegion(env, jbuf, 0, val, pbuf);
		} else  {
			err = SSL_get_error(pssl, val);

			if ((err == SSL_ERROR_WANT_READ) || (err == SSL_ERROR_WANT_WRITE)) {
				/*
				 * If either of these happens, try again
				 */
				val = -err;
			} else if (err == SSL_ERROR_SYSCALL) {
				/*
				 * If it was a system call error, get the thread's errno
				 * and check for EAGAIN or EINTR that would indicate that
				 * we should try again.
				 */
				err = jthread_geterrno();
				if ((err == EAGAIN) || (err == EINTR)) {
					/* Timeout error */
					val = 0;
				} else {
					/* Must have been a real error */
				  I802_DBG( I802_DBG_DATA, "syscall error: %d\n",err );
					val = -1;
				}
			} else {
				/* Unknown error: report it as a fatal error */
			                        I802_DBG( I802_DBG_DATA, "SSL error: %d\n",err );
				val = -1;
			}
		}
		free(pbuf);
		I802_DBG(I802_DBG_DATA, "Java_com_instant802_cldc_ssl_SSL_read: returning %d\n", val);
		return (jint) val;
	}

/**
 * SSL Write.
 * @return > 0 => # bytes written, 0 => try again, -1 for fatal errors,
 * -SSL_ERROR_WANT_READ or -SSL_ERROR_WANT_WRITE for the special OpenSSL cases
 * */
JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_write(JNIEnv *env, jobject obj, jint ssl, jbyteArray jbuf, jint buflen) {
		SSL *pssl;
		unsigned char *pbuf;
		int val, err;

		pssl = (SSL *) ssl;
		i802_ssl_init();

		pbuf = (unsigned char *) malloc(buflen);
		if (NULL == pbuf) {
			throw_memory_exception(env, "SSL_write out of memory");
			return (jint) -1;
		}
		(*env)->GetByteArrayRegion(env, jbuf, 0, buflen, pbuf);
		ERR_clear_error();
		I802_DBG(I802_DBG_DATA, "Java_com_instant802_cldc_ssl_SSL_write: calling SSL_write\n");
		val = SSL_write(pssl, pbuf, buflen);
		I802_DBG(I802_DBG_DATA, "Java_com_instant802_cldc_ssl_SSL_write: SSL_write returned %d\n", val);
		if  ( val <= 0) {
			err = SSL_get_error(pssl, val);

			if ((err == SSL_ERROR_WANT_READ) || (err == SSL_ERROR_WANT_WRITE)) {
				/*
				 * If either of these happens, try again
				 */
				val = -err;
			} else if (err == SSL_ERROR_SYSCALL) {
				/*
				 * If it was a system call error, get the thread's errno
				 * and check for EAGAIN or EINTR that would indicate that
				 * we should try again.
				 */
				err = jthread_geterrno();
				if ((err == EAGAIN) || (err == EINTR)) {
					/* Timeout error; should never happen */
					val = 0;
				} else {
					/* Must have been a real error */
					val = -1;
				}
			} else {
				/* Unknown error: report it as a fatal error */
				val = -1;
			}
		}
		free(pbuf);
		I802_DBG(I802_DBG_DATA, "Java_com_instant802_cldc_ssl_SSL_write: returning %d\n", val);
		return (jint)val;
	}

JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_useRSAPrivateKey(JNIEnv *env, jobject obj, jint ssl, jbyteArray jbuf, jint buflen, jint type) {
		SSL *pssl;
		char *fname;
		jint ret;

		pssl = (SSL *) ssl;

		i802_ssl_init();

		fname = (char *) malloc(buflen + 1);
		if (NULL == fname) {
			throw_memory_exception(env, "SSL_useRSAPrivateKeyout of memory");
			return (jint) -1;
		}
		(*env)->GetByteArrayRegion(env, jbuf, 0, buflen, fname);
		fname[buflen] = 0; //  null term "C" strings
		ret =  (jint)(SSL_use_PrivateKey_file(pssl, fname, type));
		free(fname);
		return ret;
	}

JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_useCertificateFile(JNIEnv *env, jobject obj, jint ssl, jbyteArray jbuf, jint buflen, jint type) {
		SSL *pssl;
		char *fname;
		jint ret;

		pssl = (SSL *) ssl;

		i802_ssl_init();

		fname = (char *) malloc(buflen + 1);
		if (NULL == fname) {
			throw_memory_exception(env, "SSL_useCertificateFile out of memory");
			return (jint) -1;
		}
		(*env)->GetByteArrayRegion(env, jbuf, 0, buflen, fname);
		fname[buflen] = 0; //  null term "C" strings

		ret =  (jint)(SSL_use_certificate_file(pssl, fname, type));
		free(fname);
		return ret;
	}

JNIEXPORT
void
	Java_com_instant802_cldc_ssl_SSL_close(JNIEnv *env, jobject obj, jint ssl) {
		SSL *pssl;

		pssl = (SSL *) ssl;
		//SSL_shutdown(pssl);
		SSL_free(pssl);
	}

JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_getMasterKeySize(JNIEnv *env, jobject obj, jint ssl) {
		SSL *pssl;
		SSL_SESSION *pssls;

		pssl = (SSL *)ssl;
		pssls = SSL_get_session(pssl);
		return (jint) (pssls->master_key_length);
	}

JNIEXPORT
void
	Java_com_instant802_cldc_ssl_SSL_getMasterKey(JNIEnv *env, jobject obj, jint ssl, jbyteArray jbuf, jint buflen) {
		SSL *pssl;
		SSL_SESSION *pssls;

		pssl = (SSL *)ssl;
		pssls = SSL_get_session(pssl);

		(*env)->SetByteArrayRegion(env, jbuf, 0, buflen, pssls->master_key);
	}

JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_getRandomSize() {

		return (jint)(SSL3_RANDOM_SIZE);
	}

JNIEXPORT
void
	Java_com_instant802_cldc_ssl_SSL_getClientRandom(JNIEnv *env, jobject obj, jint ssl, jbyteArray jbuf, jint buflen) {
		SSL *pssl;
		SSL_SESSION *pssls;

		pssl = (SSL *)ssl;
		pssls = SSL_get_session(pssl);

		(*env)->SetByteArrayRegion(env, jbuf, 0, buflen, pssl->s3->client_random);
	}

JNIEXPORT
void
	Java_com_instant802_cldc_ssl_SSL_getServerRandom(JNIEnv *env, jobject obj, jint ssl, jbyteArray jbuf, jint buflen) {
		SSL *pssl;
		SSL_SESSION *pssls;

		pssl = (SSL *)ssl;
		pssls = SSL_get_session(pssl);

		(*env)->SetByteArrayRegion(env, jbuf, 0, buflen, pssl->s3->server_random);
	}

/*
 * TLS PRF from RFC 2246
 */
static void P_hash(const EVP_MD *evp_md,
		   const unsigned char *secret, unsigned int secret_len,
		   const unsigned char *seed,   unsigned int seed_len,
		   unsigned char *out, unsigned int out_len)
{
	HMAC_CTX ctx_a, ctx_out;
	unsigned char a[HMAC_MAX_MD_CBLOCK];
	unsigned int size;

	HMAC_CTX_init(&ctx_a);
	HMAC_CTX_init(&ctx_out);
	HMAC_Init_ex(&ctx_a, secret, secret_len, evp_md, NULL);
	HMAC_Init_ex(&ctx_out, secret, secret_len, evp_md, NULL);

	size = HMAC_size(&ctx_out);

	/* Calculate A(1) */
	HMAC_Update(&ctx_a, seed, seed_len);
	HMAC_Final(&ctx_a, a, NULL);

	while (1) {
		/* Calculate next part of output */
		HMAC_Update(&ctx_out, a, size);
		HMAC_Update(&ctx_out, seed, seed_len);

		/* Check if last part */
		if (out_len < size) {
			HMAC_Final(&ctx_out, a, NULL);
			memcpy(out, a, out_len);
			break;
		}

		/* Place digest in output buffer */
		HMAC_Final(&ctx_out, out, NULL);
		HMAC_Init_ex(&ctx_out, NULL, 0, NULL, NULL);
		out += size;
		out_len -= size;

		/* Calculate next A(i) */
		HMAC_Init_ex(&ctx_a, NULL, 0, NULL, NULL);
		HMAC_Update(&ctx_a, a, size);
		HMAC_Final(&ctx_a, a, NULL);
	}

	HMAC_CTX_cleanup(&ctx_a);
	HMAC_CTX_cleanup(&ctx_out);
	memset(a, 0, sizeof(a));
}

static void PRF(const unsigned char *secret, unsigned int secret_len,
		const unsigned char *seed,   unsigned int seed_len,
		unsigned char *out, unsigned char *buf, unsigned int out_len)
{
        unsigned int i;
        unsigned int len = (secret_len + 1) / 2;
	const unsigned char *s1 = secret;
	const unsigned char *s2 = secret + (secret_len - len);

	P_hash(EVP_md5(),  s1, len, seed, seed_len, out, out_len);
	P_hash(EVP_sha1(), s2, len, seed, seed_len, buf, out_len);

	for (i=0; i < out_len; i++) {
	        out[i] ^= buf[i];
	}
}

#define EAPTLS_PRF_LABEL        "client EAP encryption"
#define LABEL_SIZE				21
#define SEED_SIZE 				(LABEL_SIZE + 2 * SSL3_RANDOM_SIZE)
#define EAPTLS_MPPE_KEY_LEN     32

/*
 * Generate keys according to RFC 2716
 */
JNIEXPORT
void
	Java_com_instant802_cldc_ssl_SSL_getRecvSendKeys(JNIEnv *env, jobject obj, jint ssl, jbyteArray recvbuf, jbyteArray sendbuf) {
		// (ssl, byte recv[32], byte send[32])
		SSL *s;
		unsigned char out[2*EAPTLS_MPPE_KEY_LEN], buf[2*EAPTLS_MPPE_KEY_LEN];
		// unsigned char seed[sizeof(EAPTLS_PRF_LABEL)-1 + 2*SSL3_RANDOM_SIZE];
		unsigned char seed[SEED_SIZE];
		unsigned char *p = seed;
		s = (SSL *) ssl;

		memcpy(p, EAPTLS_PRF_LABEL, LABEL_SIZE);
		p += LABEL_SIZE;
		memcpy(p, s->s3->client_random, SSL3_RANDOM_SIZE);
		p += SSL3_RANDOM_SIZE;
		memcpy(p, s->s3->server_random, SSL3_RANDOM_SIZE);

		PRF(s->session->master_key, s->session->master_key_length,
	    	seed, SEED_SIZE, out, buf, 2 * EAPTLS_MPPE_KEY_LEN);

		p = out;
		(*env)->SetByteArrayRegion(env, recvbuf, 0, EAPTLS_MPPE_KEY_LEN, p);
		p += EAPTLS_MPPE_KEY_LEN;
		(*env)->SetByteArrayRegion(env, sendbuf, 0, EAPTLS_MPPE_KEY_LEN, p);
}

/*
 * Error mapping
 */
JNIEXPORT
jint
	Java_com_instant802_cldc_ssl_SSL_getError(JNIEnv *env, jobject obj, jint ssl, jint retcode) {
		// (int ssl, int retcode)
		SSL *s;

		s = (SSL *) ssl;

		i802_ssl_init();

		return (jint)(SSL_get_error(s, retcode));
}

/*
* This doesn't belong here but we are late and it needs to go somewhere
* that won't impact the build ..... sigh ...
*/
void throw_memory_exception(JNIEnv *env, char *msg) {
	jclass clazz;

	clazz = (*env)->FindClass(env, "java/lang/OutOfMemoryError");
	(*env)->ThrowNew(env, clazz, msg);
}
