/**
 * apc_input.h - Declarations for validation.
 * Copyright: Copyright 2004 Instant802 Networks, Inc. All Rights Reserved.
 */

#include <stdlib.h>
#define __USE_GNU 1
#include <string.h>

/* holds a list of valid ints/chars for comparison */
struct apc_input_set {
	int *ints; /* array of integers.
		    * the integers may be representing chars
		    */
	int size;  /* number of integers */
};

/*
 * Destroys a struct apc_input_set.
 * Args: the struct apc_input_set.
 */
void apc_input_free_set(struct apc_input_set *set);

/* Generates a struct apc_input_set from a given range
 * ie. given 147 and 151, you'll get a struct apc_input_set whose members
 * are 147, 148, 149, 150, and 151.
 *
 * if an existing struct apc_input_set is given, it'll be merged into the
 * new set. the existing set will NOT be automatically destroyed.
 *
 * Args: min
 *       max
 *       an existing struct apc_input_set
 * Returns: a new struct apc_input_set
 */
struct apc_input_set *apc_input_get_set_from_range(int min, int max,
					struct apc_input_set *existing_set);

/* Generates a struct apc_input_set from a given string.
 * This helps bridge the gap from chars to the array of ints that's stored
 * in a struct apc_input_set.
 * Args: the string
 *       the length of the string
 * Returns: a new struct apc_input_set
 */
struct apc_input_set *apc_input_new_string_set(const char *string, int len);

/************* INPUT CHECKS *****************/
/* The rest of these are input checking functions. They return 0 if
 * false and non-zero if true.
 * They always take at least two arguments: the string your want to check and
 * the length of that string.
 */

/* Is it a MAC Address? */
int apc_input_is_mac(const char *string, int len);

/* Is it an IP Address? */
int apc_input_is_ip(const char *string, int len);

/* Is it an integer? */
int apc_input_is_int(const char *string, int len);

/* Is it an integer that belongs in a given set?
 * Additional args: a struct apc_input_set
 */
int apc_input_is_int_in_set(const char *string, int len, struct apc_input_set *valid_ints);

/* Is it a positive integer? */
int apc_input_is_positive_int(const char *string, int len);

/* Is it zero? */
int apc_input_is_zero(const char *string, int len);

/* Is it a non-negative integer? */
int apc_input_is_non_negative_int(const char *string, int len);

/* Is it a negative integer? */
int apc_input_is_negative_int(const char *string, int len);

/* Is it an integer that's within a range?
 * Additional args: min
 *                  max
 */
int apc_input_is_int_in_range(const char *string, int len, int min, int max);

int apc_input_is_percent(const char *string, int len);

int apc_input_is_even(const char *string, int len);

int apc_input_is_double_in_range(const char *string, int len,
				 double min, double max);

/* Is it a string of chars that all belong to a given set?
 * Additional args: a struct apc_input_set
 */
int apc_input_is_string_in_set(const char *candidate, int len,
				struct apc_input_set *valid_chars);

/* Is it an integer that's within an ASCII range?
 * Additional args: min
 *                  max
 */
int apc_input_is_string_in_range(const char *string, int len, char min, char max);


/* Is it an integer that's within a collection of ASCII range?
 * Additional args: valid ranges
 *                    e.g. "az09" is a collection of two ASCII ranges:
 *                       the ASCII range from 'a' to 'z' and the ASCII
 *                       range from '0' to '9'.
 *                  number of valid ranges
 *                       usually just strlen(ranges)/2
 */
int apc_input_is_string_in_ranges(const char *string, int len,
				  const char *valid_ranges,
				  int num_ranges);

int apc_input_is_string_in_array(const char *string, int len,
				 const char * const array[]);

/* Is it a string whose chars belong in the ranges "09AZaz--''"  ? */
int apc_input_is_word(const char *string, int len);

/* Is it a string whose chars belong in the ranges "09AZaz,-\t\t  ')" ? */
int apc_input_is_phrase(const char *string, int len);

/* Is it a valid URI? */
int apc_input_is_uri(const char *string, int len);

/* Is the string a valid ssid? */
int apc_input_is_ssid(const char *string, int len);

int apc_input_is_mode(const char *string, int len);

int apc_input_is_any_channel(const char *string, int len);

/* Regulatory domains. */
typedef	enum {
	APC_REG_FCC    = 0,
	APC_REG_ETSI   = 1,
	APC_REG_TELEC  = 2
} apc_reg;

int apc_input_is_a_channel(const char *string, int len, apc_reg reg);
int apc_input_is_a_rate(const char *string, int len);

int apc_input_is_b_channel(const char *string, int len, apc_reg reg);
int apc_input_is_b_rate(const char *string, int len);

int apc_input_is_g_channel(const char *string, int len, apc_reg reg);
int apc_input_is_g_rate(const char *string, int len);

int apc_input_is_turbo_a_channel(const char *string, int len, apc_reg reg);
int apc_input_is_turbo_a_rate(const char *string, int len);

int apc_input_is_dynamic_turbo_a_channel(const char *string, int len,
					 apc_reg reg);
int apc_input_is_dynamic_turbo_a_rate(const char *string, int len);

int apc_input_is_turbo_g_channel(const char *string, int len, apc_reg reg);
int apc_input_is_turbo_g_rate(const char *string, int len);

int apc_input_is_dynamic_turbo_g_channel(const char *string, int len,
					 apc_reg reg);
int apc_input_is_dynamic_turbo_g_rate(const char *string, int len);

int apc_input_is_channel(const char *string, int len,
			 const char *mode, apc_reg reg);

int apc_input_is_domain_label(const char *string, int len);

int apc_input_is_domain_name(const char *string, int len);

int apc_input_is_ip_or_domain_name(const char *string, int len);

/* Is it "up" or "down"? */
int apc_input_is_up_down(const char *string, int len);

/* Is it "up" or "down"? */
int apc_input_is_on_off(const char *string, int len);

int apc_input_is_yes_no(const char *string, int len);

int apc_input_is_interface(const char *string, int len);

/* Is it a WPA PSK key? */
int apc_input_is_wpa_psk_key(const char *string, int len);

int apc_input_is_wep_key(const char *string, int len,
			 int key_length, int ascii);

int apc_input_is_wep_key_length(const char *string, int len);

int apc_input_is_regulatory_domain(const char *string, int len);

int apc_input_is_string_in_array(const char *string,
				 int len,
				 const char * const array[]);
