/*
 * Created on Oct 29, 2003
 * author: Robin Barooah
 * (C) 2003 Instant802 Networks Inc
 */
package com.instant802.cldc.clustermastering;

import com.instant802.cldc.coordination.CoordinationAgent;
import com.instant802.cldc.deviceconfiguration.ElementNotFoundException;
import com.instant802.cldc.net.SocketException;
import com.instant802.cldc.orchestrator.OrchestratorNode;
import com.instant802.cldc.util.DateUtil;
import com.instant802.cldc.util.Log;
import com.instant802.cldc.util.ThreadCounter;

/**
 * @author robin
 */
public class ClusterMasterAgent2 implements ClusterMasterAgent {

	private final static long SETTLE_TIME = 15000;
	private final static long REBROADCAST_INTERVAL = 2500;
	private final static long MAX_PERTURBATION = 500;
	private final static long MASTER_CONFIRMATION_INTERVAL = 45000;

	private long nextMasterAnnouncement = 0;	
	private ClusterMaster currentMaster;
	ClusterMaster prospectiveMaster;
	private OrchestratorNode node;
	private ElectionMonitor electionMonitor;
	boolean active = true;

	public ClusterMasterAgent2(OrchestratorNode node) {
		this.node = node;
	}

	public long intervalPerterbation() {
		long now = DateUtil.now();
		return now % MAX_PERTURBATION;
	}

	public synchronized ClusterMaster getMaster()
		throws
			ElementNotFoundException,
			InvalidTimingException,
			SocketException,
			AgentNotRunningException {
		if (null == currentMaster) {
			Log.info("null is current master - so starting election");
			startElection();
		}
		while (!masterSettled()) {
			try {
				wait(masterTimeRemaining());
			} catch (InterruptedException e) {
				error(e);
			}
		}
		return currentMaster;
	}

	/**
	 * this method is called by other objects when they detect that
	 * the current master is invalid.
	 */
	public synchronized void invalidateMaster()
		throws ElementNotFoundException, SocketException {
		currentMaster = null;
		// Don't need to start an election, as this will happen
		// if we need a master later.
//		if (!electing()) {
//			startElection();
//		}
	}

	private synchronized void startElection()
		throws ElementNotFoundException, SocketException {
		if (!electing()) {
			Log.info("election isn't in progress so sending broadcast");
			AnnounceMasterMessage message = new AnnounceMasterMessage(node);
			node.broadcastMessage(message);
			if (prospectiveMaster == null) {
				setProspective(new ClusterMaster(node, message));
			}
			startMonitor();
		} else {
			Log.info("election already in progress");
			// already electing, so not starting monitor or sending broadcast
		}
	}

	void error(Exception e) {
		Log.error("ClusterMasterAgent2", e);
	}

	/** 
	 * return true if the current master is settled.
	 */
	private synchronized boolean masterSettled() {
		if (null == currentMaster) {
			return false;
		}
		boolean settled = currentMaster.age() > SETTLE_TIME;
		return settled;
	}

	/**
	 * this method is called whenever an announcement message is received.
	 */
	public synchronized void handleMasterAnnouncement(AnnounceMasterMessage message)
		throws ElementNotFoundException, SocketException {
		if (participating()) {
			considerCandidate(message);
		} else {
			// we're not participating so ignoring the announcement
		}
	}

	/**
	 * announce ourselves as a potential master.  Don't start an election from our point
	 * of view unless we receive a response from a superior master.
	 */
	public void makeAnnouncement()
		throws ElementNotFoundException, SocketException {
		node.broadcastMessage(new AnnounceMasterMessage(node));
	}

	/**
	 * consider a message from a potential candidate.
	 */
	private synchronized void considerCandidate(AnnounceMasterMessage message)
		throws ElementNotFoundException, SocketException {
		if (null == prospectiveMaster) {
			setProspective(
				new ClusterMaster(node, new AnnounceMasterMessage(node)));
		}
		if (!electing()) {
			try {
				startMonitor();
			} catch (Exception e) {
				error(e);
			}
		}
		/* if the new message comes from the existing prospective master, 
		 * we don't replace the existing record.  This is to prevent thrashing
		 * as the precedence changes.
		 */
		
		if( ! prospectiveMaster.sentMessage( message ) )
		{		
			long score = criteria().messageScore(message);
			if (score > prospectiveMaster.score()) {
				setProspective(new ClusterMaster(node, message));
			} else if (prospectiveMaster.isLocal()) {
					node.broadcastMessage(new AnnounceMasterMessage(node));
			}
		}
	}

	private MasterCriteria criteria() {
		return node.masterCriteria();
	}

	synchronized void setProspective(ClusterMaster master) {
		prospectiveMaster = master;
	}

	public void deactivate() {
		active = false;
	}

	public void activate() {
		active = true;
	}

	/** return true if the prospective master is settled 
	 *
	 */
	synchronized boolean prospectiveSettled() {
		return (null != prospectiveMaster)
			? prospectiveMaster.age() > SETTLE_TIME
			: false;
	}

	/**
	 * return the number of milliseconds remaining before the current
	 * prospective master would be settled.
	 */
	long prospectiveTimeRemaining() {
		return (null != prospectiveMaster)
			? SETTLE_TIME - prospectiveMaster.age()
			: SETTLE_TIME;
	}

	/**
	 * return the number of milliseconds remaining before the current
	 * actual master would be settled.
	 */
	long masterTimeRemaining() {
		return (null != currentMaster)
			? SETTLE_TIME - currentMaster.age()
			: SETTLE_TIME;
	}

	/**
	 * return true if the agent should participate in master elections.
	 * if we are are currently a client, we should not participate
	 * because this means we lost a previous election to the server to whom
	 * we are connected.  We should wait until this server becomes dislodged.
	 */
	private boolean participating() {
		boolean client = coordinationAgent().isClient();
		boolean participating = active && (!client);
		return participating;
	}

	/**
	 * make the prospective master into the current master
	 */
	synchronized void electProspective() {
		if (currentMaster != null) {
			if (currentMaster.getId() != prospectiveMaster.getId()) {
				acceptProspective();
				masterChanged();
			}
		} else {
			acceptProspective();
			masterChanged();
		}
		notifyAll();
	}

	private void acceptProspective() {
		currentMaster = prospectiveMaster;
		prospectiveMaster = null;
	}

	synchronized boolean electing() {
		return electionMonitor != null;
	}

	synchronized void setMonitor(ElectionMonitor monitor) {
		electionMonitor = monitor;
	}

	synchronized void startMonitor() {
		if (null == electionMonitor) {
			electionMonitor = new ElectionMonitor(this);
			electionMonitor.start();
		} else {
			// monitor exists, so not creating a new one
		}
	}

	/**
	 *  notify the listener that the master has changed 
	 */
	private void masterChanged() {
		coordinationAgent().masterChanged();
	}

	private CoordinationAgent coordinationAgent() {
		return node.coordinationAgent();
	}

	private class ElectionMonitor extends Thread {

		private final ClusterMasterAgent agent;

		ElectionMonitor(ClusterMasterAgent agent) {
			this.agent = agent;
			ThreadCounter.register(this);
		}

		public void run() {
			synchronized (agent) {
				try {
					while (!prospectiveSettled()) {
						readvertize();
						try {
							long duration = intervalPerterbation() + REBROADCAST_INTERVAL;
							agent.wait(duration);
						} catch (InterruptedException e) {
							error(e);
						}
					}
					if (active) {
						electProspective();
					} else {
						setProspective( null );
					}
				} finally {
					setMonitor(null);
				}
			}
		}

		private void readvertize() {
			if( prospectiveMaster.isLocal() ) {
				try {
					makeAnnouncement();
				} catch (ElementNotFoundException e1) {
					error( e1 );
				} catch (SocketException e1) {
					error( e1 );
				}
			}
		}

		public boolean electing() {
			return isAlive();
		}
	}
	public synchronized void confirmMaster() {
		if( DateUtil.now() > nextMasterAnnouncement ) {
			nextMasterAnnouncement = DateUtil.now() + MASTER_CONFIRMATION_INTERVAL;
			try {
				makeAnnouncement();
			} catch (ElementNotFoundException e) {
				error( e );
			} catch (SocketException e) {
				error( e );
			}
		}
	}
	
	public ClusterMaster currentMaster() {
		return currentMaster;
	}
}
