/*
 * Created on Oct 5, 2003
 * written by robin
 * (C) 2003 Instant802 Networks Inc
 */
package com.instant802.cldc.clustermembership;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import com.instant802.cldc.DataPath.XMLRepresentationException;
import com.instant802.cldc.LSQL.APConfException;
import com.instant802.cldc.LSQL.LSQLException;
import com.instant802.cldc.clustermastering.MasterCriteria;
import com.instant802.cldc.deviceconfiguration.ClusteredAP;
import com.instant802.cldc.deviceconfiguration.ConfigurationModel;
import com.instant802.cldc.net.MalformedHTTPMessageException;
import com.instant802.cldc.net.SocketException;
import com.instant802.cldc.util.Log;
import com.instant802.cldc.util.SimpleCommaDelimParser;

/**
 * @author robin
 */
public class MembershipList {

	private final MasterCriteria criteria;
	private final ConfigurationModel model;
	private final Hashtable table = new Hashtable();

	public MembershipList(MasterCriteria criteria, ConfigurationModel model) {
		this.criteria = criteria;
		this.model = model;
	}

	/**
	 * populate the membership list by clearing it and reading
	 * the contents from the configuration model.
	 */
	public synchronized void populate() {
		synchronized (model) {
			Enumeration e = model.getClusteredAPs();
			table.clear();
			while (e.hasMoreElements()) {
				ClusteredAP ap = (ClusteredAP) e.nextElement();
				table.put(ap.id(), ap );
//				table.put(ap.id(), new StoredClusterMember(ap));
			}
			updateCriteria();
		}
	}

	private final void updateCriteria() {
		criteria.setMembers( table.size() );
	}

	/**
	 * add or update the membership list to include the member
	 * as required.
	 */
	public synchronized void addMember(ReceivedClusterMember member)
		throws
			LSQLException, APConfException, XMLRepresentationException, IOException, MalformedHTTPMessageException, SocketException {
		synchronized (model) {
			ClusteredAP found = (ClusteredAP) table.get(member.id());
			if (null == found) {
				if (Log.DEBUG) debug("member is not already in list, so adding it");
				add(member);
			} else {
				if (Log.DEBUG) debug("member is already in the list");
				if (!found.ipAddress().equals(member.ipAddress())) {
					info("member IP address has changed - so updating the list");
					update(member);
				}
			}
		}
	}

	private void info(String string) {
		Log.info("MembershipList", string );
	}

	private void debug(String string) {
		Log.debug("MembershipList", string);
	}

	/**
	 * update the details of a member in both the membership list
	 * and the configuration model, and distribute the configuration
	 * to the members.
	 */
	private synchronized void update(ReceivedClusterMember member)
		throws
			LSQLException, APConfException, XMLRepresentationException, IOException, MalformedHTTPMessageException, SocketException {
		ClusteredAP apRecord = new ClusteredAP( model.schema(), member ); 
		table.put(member.id(), apRecord );
		Enumeration e = model.getClusteredAPs();
		while (e.hasMoreElements()) {
			ClusteredAP ap = (ClusteredAP) e.nextElement();
			if (ap.id().equals(member.id())) {
				if (!ap.ipAddress().equals(member.ipAddress())) {
					if (Log.DEBUG) debug("found matching record - changing IP address in record");
					ap.ipAddress(member.ipAddress());
					if (Log.DEBUG) debug("writing configuration back to apconfD");
					model.writeToAPConf();
					model.distributeConfiguration();
				}
			}
		}
		updateCriteria();
	}

	/**
	 * add a new member to the membership list and the configuration model.
	 */
	private synchronized void add(ReceivedClusterMember member)
		throws
			LSQLException, APConfException, XMLRepresentationException, IOException, MalformedHTTPMessageException, SocketException {
		synchronized (model) {
			if (Log.DEBUG) debug("adding new cluster member to list and configuration model");
			ClusteredAP ap = new ClusteredAP( model.schema(), member ); 
			if (Log.DEBUG) debug("adding member to internal table");
			table.put( member.id(), ap	);
			if (Log.DEBUG) debug("adding member to configuration model");
			model.addObject( ap );
			if (Log.DEBUG) debug("updating criteria");
			updateCriteria();
			if (Log.DEBUG) debug("writing configuration back to apconfD");
			model.writeToAPConf();
			if (Log.DEBUG) debug("distributing configuration");
			model.distributeConfiguration();
		}
	}

	/**
	 * return the current number of members.
	 */
	public synchronized int size() {
		synchronized (model) {
			return table.size();
		}
	}

	/**
	 * return an enumeration over a snapshot of the current 
	 * set of members.  Once acquired, this enumeration will
	 * not be affected if the membership list is updated.
	 */
	public synchronized Enumeration elements() {
		Vector v = new Vector();
		Enumeration e = table.elements();
		while( e.hasMoreElements() ) {
			v.addElement( new StoredClusterMember( (ClusteredAP) e.nextElement() ) );
		}
		if (Log.DEBUG) debug("returning element enumeration with "+v.size()+" members");
		return v.elements();
	}

	/**
	 * remove a single member from both the membership list and the configuration model.
	 */
	synchronized void removeMember(ClusteredAP ap) throws LSQLException, APConfException {
		synchronized( model ) {
			table.remove( ap.id() );
			model.deleteObject( ap );
			updateCriteria();
		}
	}

	/**
	 * remove all members from the membership list and the configuration model.
	 */
	synchronized void removeAll() throws LSQLException, APConfException {
		synchronized( model ) {
			Enumeration e = model.getClusteredAPs();
			Vector toRemove = new Vector();
			while( e.hasMoreElements() ) {
				toRemove.addElement( e.nextElement() );
			}
			
			e = toRemove.elements();
			while( e.hasMoreElements() ) {
				ClusteredAP ap = (ClusteredAP) e.nextElement();
				model.deleteObject( ap );
			}
			table.clear();
			updateCriteria();
		}
	}

	public synchronized boolean containsMember(MembershipRequestMessage message) {
		return table.containsKey( message.getSenderId() );
	}
	
	public synchronized boolean containsMember(String wirelessmac) {
		Enumeration apsen = table.elements();
		while (apsen.hasMoreElements()) {
			ClusteredAP ap = (ClusteredAP)apsen.nextElement();
			if (containsMac(ap.radioMacs(), wirelessmac))
				return true;
		}
		return false;
	}	

	public synchronized boolean containsMemberIP(String ip) {
		Enumeration apsen = table.elements();
		while (apsen.hasMoreElements()) {
			ClusteredAP ap = (ClusteredAP)apsen.nextElement();
			if (ap.ipAddress().equals(ip))
				return true;
		}
		return false;		
	}
	
	private boolean containsMac(String list, String addr) {
		SimpleCommaDelimParser scdp = new SimpleCommaDelimParser(list);
		String mac;
		Vector v = new Vector();
		while (null != (mac = scdp.nextField())) {
			if (mac.equals(addr))
				return true;
		}
		return false;
	}
}
