package com.instant802.cldc.coordination;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import com.instant802.cldc.DataPath.XMLRepresentationException;
import com.instant802.cldc.LSQL.APConfException;
import com.instant802.cldc.LSQL.LSQLException;
import com.instant802.cldc.channelplanner.WirelessNeighborhoodRecords;
import com.instant802.cldc.clustermastering.AgentNotRunningException;
import com.instant802.cldc.clustermastering.InvalidTimingException;
import com.instant802.cldc.clustermembership.MarkedForRemoval;
import com.instant802.cldc.deviceconfiguration.ClusteredAP;
import com.instant802.cldc.deviceconfiguration.ElementNotFoundException;
import com.instant802.cldc.messaging.Message;
import com.instant802.cldc.net.MalformedHTTPMessageException;
import com.instant802.cldc.net.SocketException;
import com.instant802.cldc.net.SocketTimeoutException;
import com.instant802.cldc.orchestrator.OrchestratorNode;
import com.instant802.cldc.synchronization.SynchronizationSetMessage;
import com.instant802.cldc.transfer.TransferBase;
import com.instant802.cldc.util.Log;

/**
 * @author robin
 *
 * To change the template for this generated type comment go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
public class CoordinatorServer extends TransferBase implements Coordinator {

	protected final OrchestratorNode node;
	private final Hashtable agents = new Hashtable();
	boolean running = true;
	private final Boolean block = new Boolean(true);
	private static final int SENDERS_MAX = 3;
	
	public CoordinatorServer(OrchestratorNode node) throws SocketException {
		super(
			node.schema(),
			node.sslSessionContext(),
			node.getCoordinationPort(),
			node.adminInterface() );
		this.node = node;
	}

	public void coordinate()
		throws
			ElementNotFoundException,
			SocketException,
			XMLRepresentationException,
			IOException,
			MalformedHTTPMessageException,
			LSQLException,
			APConfException {
		if (Log.DEBUG) debug("starting to coordinate");
		try {
			new WirelessNeighborhoodRecords(node).startAutoGather();
		} catch (Exception e){
			error(e);
		}
		try {
			node.membershipAgent().controlMembership();
			run();
		} finally {
			WirelessNeighborhoodRecords wnr = WirelessNeighborhoodRecords.getWirelessNeighborhoodRecords();
			if (null!= wnr)
				wnr.stopAutoGather();
		}
		if (Log.DEBUG) debug("ending coordination");
	}

	void debug(String message) {
		Log.debug("CoordinatorServer", message);
	}
	
	protected void receiveConnection(ClientAgent agent) {
		if (Log.DEBUG) Log.info("received incoming coordination connection");
		agent.start();
	}

	public void addAgent(ClientAgent agent) {
		if (Log.DEBUG) Log.notice("approved incoming coordination connection");
		synchronized (agents) {
			agents.put( agent.id(), agent);
		}
	}

	public void dropAgent(ClientAgent agent) {
		if (Log.DEBUG) Log.info("dropping failed coordination connection");
		synchronized (agents) {
			agents.remove(agent.id());
			agent.close();
		}
	}

	private Vector getAgents() {
	    Vector v = new Vector();
	    synchronized (agents) {
		    Enumeration e = agents.elements();
		    while( e.hasMoreElements() ) {
		    	v.addElement( e.nextElement() );
		    }
	    }
	    return v;
	}

	public void stopCoordinating() {
		try{
		running = false;
		if (Log.DEBUG) debug("closing sockets");
		Vector v;
	        close();
	        if (Log.DEBUG) debug("getting list of agents");
		v = getAgents();
		// v becomes a snapshot of the agents that were registered when we close
		// the server socket.  Because the server socket is closed, no new agents
		// can be accepted.
		if (Log.DEBUG) debug("terminating agents");
		terminateAllAgents( v );
		if (Log.DEBUG) debug("done terminating agents");
		} catch ( RuntimeException e ) { 
			error( e );
		}
	}

	public void handleMessage(MessageHolder received)
		throws
			ElementNotFoundException, XMLRepresentationException, IOException, MalformedHTTPMessageException, LSQLException, APConfException, InvalidTimingException, SocketException, AgentNotRunningException {
		if (Log.DEBUG) Log.info("handling message " + received.getMessage().typeName()+" from "+received.getMessage().getSenderIp() );

		synchronized (node.configurationModel()) {
			received.getMessage().onReceipt(node);
			if (received.getMessage().typeName() == SynchronizationSetMessage.TYPE) {
				received.clear();
				broadcastConfiguration();
				handleRemovedAPs();
				statusConnected();
			}
		}
	}

	private void broadcastConfiguration() throws ElementNotFoundException {
		node.nodeStatusRecord().setClusterStatus("distributing configuration change to cluster");
		sendToAll(
			new SynchronizationSetMessage(
				node,
				node.configurationModel().currentObjects()) );
		
	}

	/** check whether any aps have been deleted and act accordingly.
	 * 
	 */
	private void handleRemovedAPs() throws LSQLException, APConfException, XMLRepresentationException, IOException, MalformedHTTPMessageException, SocketException {
		// 1. see if aps are marked removed.
		// 2. delete them from the configuration
		// 3. send a configuration update.
		// 4. if the coordinator was one of the aps, drop into standalone mode.
		MarkedForRemoval marked = node.membershipAgent().getMarkedForRemoval();
		if (Log.DEBUG) Log.info("handling removed APs");
		if( marked.notEmpty() ) {
			if (Log.DEBUG) Log.info("set marked is not empty");
			marked.deleteFromConfig();
			node.configurationModel().writeToAPConf();
			terminateAgents( marked );
			broadcastConfiguration();
			if( marked.includesThis() ) {
				setClusterStatus("removed from cluster - going into standalone mode...");				
				node.enterStandalone();
			}
		}
	}

	/**
	 * abruptly terminate agents corresponding to access points which have been marked for removal
	 * and remove them from the list of agents to broadcast to.
	 * @param marked the set of aps marked for removal.
	 */
	private void terminateAgents(MarkedForRemoval marked) {
		Enumeration e = marked.elements();
		synchronized (agents) {
			while( e.hasMoreElements() ) {
				ClusteredAP ap = (ClusteredAP) e.nextElement();
				Object o = agents.remove( ap.id() );
				try{
					ClientAgent removed = (ClientAgent) o; 
					if( null != removed ) {
						removed.close();
					} else {
						// removed ap must be self, so no agent to remove
					} 
				} catch ( ClassCastException f ) {
					reportClassCast( o , f );
				}
			}
		}
	}

	
	private void terminateAllAgents( Vector v ) {
		terminateAgents( v );
		waitForAgentsToStop( v );
		agents.clear();
	}

	private void waitForAgentsToStop( Vector v ) {
		Enumeration e = v.elements();
		while( e.hasMoreElements() ) {
			Object o  = e.nextElement();
			try{			
				ClientAgent agent = (ClientAgent) o;
				if (Log.DEBUG) info("waiting for agent to stop");
				agent.waitForStop();
				if (Log.DEBUG) info("done waiting for agent to stop");
			} catch ( ClassCastException f ) {
				reportClassCast(o, f);
			}
		}		
	}

	private void reportClassCast(Object o, ClassCastException f) {
		error( f  );
		if (Log.DEBUG) debug("couldn't cast "+o+" to ClientAgent");
	}

	private void terminateAgents( Vector v ) {
		Enumeration e = v.elements();
		while( e.hasMoreElements() ) {
			Object o = e.nextElement();
			try{
				ClientAgent agent = (ClientAgent) o;
				agent.close();
			} catch ( ClassCastException f ) {
				reportClassCast( o, f );
			}
		} 
	}

	/**
	 * broadcast a message to all of the aps in the cluster.
	 */
	public void broadcast(Message received ) throws LSQLException, APConfException, XMLRepresentationException, IOException, MalformedHTTPMessageException, SocketException {

		synchronized (node.configurationModel()) {
			sendToAll(received);
			handleRemovedAPs();
			node.lsqlMessageReceiver().clearLocallyModified();	
		}
	}

	private void sendToAll(Message received) {
		if (Log.DEBUG) Log.info("broadcasting message to " + agents.size() + " clients");
//		AgentGroup group = new AgentGroup( agents );
//		group.send( received );
		Enumeration e = agents.elements();
		StringBuffer stringMessage = received.asXMLStringBufferDocument();
		SenderPool senders = startSenders(stringMessage, received.typeName(), e);
		stringMessage = null;
//		awaitCompletion(senders);
		senders.awaitCompletion();
	}

	/**
	 * return when all theads in the list have completed.
	 */
//	private void awaitCompletion(Vector senders) {
//		Enumeration s = senders.elements();
//		while( s.hasMoreElements() ) {
//			Thread toWaitFor = (Thread) s.nextElement();
//			try {
//				toWaitFor.join();
//			} catch (InterruptedException e1) {
//				error( e1 );
//			}
//		}
//	}

	/**
	 * start a sender thread to send a message to each agent in parallel 
	 */
//	private Vector startSenders(Message received, Enumeration e ) {
//		Vector senders = new Vector();
//		while (e.hasMoreElements()) {
//			ClientAgent agent = (ClientAgent) e.nextElement();
//				Sender sender = new Sender(agent, received);
//				sender.start();
//				senders.addElement(sender);
//				//					agent.send( received );
//		}
//		return senders;
//	}

//	private Vector startSenders(StringBuffer message, String typeName, Enumeration e ) {
//		Vector senders = new Vector();
//		while (e.hasMoreElements()) {
//			ClientAgent agent = (ClientAgent) e.nextElement();
//				Sender sender = new Sender(agent, message, typeName );
//				sender.start();
//				senders.addElement(sender);
//				//					agent.send( received );
//		}
//		return senders;
//	}

	private SenderPool startSenders(StringBuffer message, String typeName, Enumeration e ) {
		SenderPool pool = new SenderPool( SENDERS_MAX );
		while (e.hasMoreElements()) {
			ClientAgent agent = (ClientAgent) e.nextElement();
				Sender sender = new Sender(agent, message, typeName );
				sender.start();
				pool.add(sender);
				sender = null;
		}
		return pool;
	}

	
	private class Sender extends Thread {
		private final ClientAgent agent;
		private final Message message;
		private final StringBuffer stringMessage;
		private final String typeName;
		
//		Sender(ClientAgent agent, Message message) {
//			this.agent = agent;
//			this.message = message;
//			this.stringMessage = null;
//			this.typeName = null;
//		}

		Sender(ClientAgent agent, StringBuffer stringMessage, String typeName ) {
			this.agent = agent;
			this.message = null;
			this.stringMessage = stringMessage;
			this.typeName = typeName;
		}
		
		public void run() {
			try {
				if (Log.DEBUG) debug("broadcasting to " + agent.id());
				if( stringMessage == null ) 
				{
					agent.send(message);
				} else {
					agent.send( stringMessage, typeName );
				}
			} catch (SocketException e) {
				error(e);
			}
		}

	}

	void error( Exception e1) {
		Log.error("CoordinatorServer", e1);
	}

	public boolean isServer() {
		return true;
	}

	public void close() {
		running = false;
		synchronized (block) {
			super.close();
		}
	}

	private void stop() {
		close();
	}

	public void run() {
		if (Log.DEBUG) Log.info("ready to receive incoming coordination connections");
		synchronized (block) {
			while (running) {
				statusConnected();
				try {
					if (Log.DEBUG) debug("MASTER / COORDINATOR");
					ClientAgent agent = super.accept(this);
					if (Log.DEBUG) debug("Received connection " + agent);
					receiveConnection(agent);
				} catch (SocketTimeoutException sto) {
					if (Log.DEBUG) debug("Accept timed out");
					if (running)
						node.clusterMasterAgent().confirmMaster();
				} catch (SocketException se) {
					if (Log.DEBUG) debug("Socket exception - stopping");
					stop();
				}
			}
		}
		if (Log.DEBUG) info("exiting run loop in coordinatorserver");
	}

	private void info(String string) {
		Log.info( "CoordinatorServer", string );
	}

	private void statusConnected() {
		node.nodeStatusRecord().setClusterStatus("connected to cluster.");
	}

	public boolean isActive() {
		return running;
	}

	public void handleProbeResponse(ProbeResponseMessage message) {
		// really do nothing here, as the server currently doesn't care about probe responses.
	}

	public void checkForDeletions() {
		// really do nothing here, as we need to broadcast the pre-deletion state before we handle deletions
		// when we are the server.
	}
	
	private void setClusterStatus( String message ) {
		node.nodeStatusRecord().setClusterStatus( message ); 
	}
}
