/*
 * Created on Jun 8, 2004
 * author: Robin Barooah
 * Copyright (C) 2004 Instant802 Networks, Inc. All Rights Reserved.
 */
package com.instant802.cldc.hostapd;

import java.util.Vector;

import com.instant802.cldc.DataPath.NameValuePair;
import com.instant802.cldc.DataPath.NameValuePairFileReader;
import com.instant802.cldc.channelplanner.NeighborAPRecord;
import com.instant802.cldc.channelplanner.WirelessInterface;
import com.instant802.cldc.linuxfstools.DirectoryReader;
import com.instant802.cldc.net.SocketException;
import com.instant802.cldc.net.UnixDatagramSocket;
import com.instant802.cldc.util.Log;

/**
 * @author robin
 * Object provides access to hostAPD via the datagram API
 */
public class HostAPD {
	
	private static final String PROC_PATH = "/proc/net/ieee80211";
	private static final String VAR_PATH = "/var";
	
	private static final String LOCAL_BASE = "orchestrator";
	private static final String GLOBAL_IFACE = "hostapd-all";
	private static final String CMD_PING = "PING"; 
	private static final String RESPONSE_PONG = "PONG";
	
	private static final int LINE_BUFFER = 1024;
	
	private int socketCount = 0;
	
	/**
	 * Local singleton instance.  There can be only one (for now). 
	 */
	private static HostAPD instance;
	
	/**
	 * Factory method returns the singleton HostAPD Api Object.  The singleton status of this
	 * object may change in future, so subsequent calls to this method should not be assumed
	 * to return the same object.  
	 */
	public synchronized static HostAPD getHostAPD() {
		if( instance == null ) {
			instance = new HostAPD();
		}
		return instance;
	}
	
	/**
	 * Check whether HostAPD is present and responding.  Return true if this is the case,
	 * false if the reply is not as expected.
	 * Otherwise throw an exception to indicate what failed.
	 */
	public boolean Ping() throws SocketException {
		String reply = requestResponse( GLOBAL_IFACE, CMD_PING );
		return reply.equals( RESPONSE_PONG );
	}
	
	String requestResponse( String iface, String request ) throws SocketException {
		UnixDatagramSocket sender = getNextSocket(); 
		sender.sendTo( request.getBytes(), iface );
		byte[] received = new byte[ LINE_BUFFER ];
		long length = sender.receiveFrom( received );
		sender.close();
		String reply = new String( received );
		reply = reply.substring( 0, (int) length );
		return reply;
	}

	/**
	 * Return a new datagram socket with an appropriate endpoint name.
	 */
	private UnixDatagramSocket getNextSocket() throws SocketException {
		return new UnixDatagramSocket( LOCAL_BASE + "-" + socketCount++ );
	}
	
	/**
	 * return a HostAPDRadio object that represents the named
	 * radio.
	 * @param name the name of the radio.
	 */
	public HostAPDRadio getRadio( String name ) {
		return new HostAPDRadio( this, name );
	}
		
	public Vector radios() {
		Vector v = new Vector();
		DirectoryReader dir = new DirectoryReader();
		if (Log.DEBUG) debug("opening directory (path "+PROC_PATH+")");
		dir.openDirectory( PROC_PATH );
		for( String name = dir.readDirectory(); name != null; name=dir.readDirectory() ) {
			if (Log.DEBUG) debug("filename "+name );
			if( ! dir.navReference( name ) ) {
				if (Log.DEBUG) debug("creating-radio (name "+name+")");
				v.addElement( new HostAPDRadio( this, name ) );
			}
		}
		if (Log.DEBUG) debug("returning-vector (elements "+v.size()+")");
		return v;
	}

	private void debug(String string) {
		Log.debug("HostAPD", string );
	}
	
	public String getRadioBand(String name) {
		NameValuePairFileReader read = new NameValuePairFileReader(PROC_PATH +"/" + name + "/config");
		NameValuePair nvp;
		
		while (read.hasMoreElements()) {
			nvp = (NameValuePair) read.nextElement();
			if (nvp.getName().equals("mode")) {
				String mode = nvp.getValue();
				char c = mode.charAt(mode.length() - 1);
				if (c == 'b' || c == 'g')
					mode = NeighboringAP.BAND_B_G;
				else if (c == 'a')
					mode = NeighboringAP.BAND_A;
				else
					mode = NeighboringAP.BAND_UNKNOWN;
				return mode;
			}
		}
		return NeighboringAP.BAND_UNKNOWN;
	}
	
	public String getSSID(String name) {
		NameValuePairFileReader read = new NameValuePairFileReader(VAR_PATH +"/hostapd.conf." + name);
		NameValuePair nvp;			
		while (read.hasMoreElements()) {
			nvp = (NameValuePair) read.nextElement();
			if (nvp.getName().equals("ssid")) {			
				return nvp.getValue();
			}
		}	
		return " ";
	}

	public String getRadioChannel(String name) {
		NameValuePairFileReader read = new NameValuePairFileReader(PROC_PATH +"/" + name + "/config");
		NameValuePair nvp;
		while (read.hasMoreElements()) {
			nvp = (NameValuePair) read.nextElement();
			if (nvp.getName().equals("channel"))
				return nvp.getValue();
		}
		return "-1";
	}
	
	public String getRadioLoad(String name)
	{
		NameValuePairFileReader read = new NameValuePairFileReader(PROC_PATH +"/" + name + "/counters");
		NameValuePair nvp;
		String nvpname;
		long framecount = 0;
		while (read.hasMoreElements()) {
			nvp = (NameValuePair) read.nextElement();
			nvpname = nvp.getName();
			if (nvpname.equals("TransmittedFragmentCount") ||  nvpname.equals("ReceivedFragmentCount"))
				framecount +=  Long.parseLong(nvp.getValue());
		}
		return Long.toString(framecount);
	}

	public String getRadioMac(String name)
	{
		return new WirelessInterface(name).getHardwareAddress();
	}	
}
