/*
 * Created on Jun 8, 2004
 * author: Robin Barooah
 * Copyright (C) 2004 Instant802 Networks, Inc. All Rights Reserved.
 */
package com.instant802.cldc.hostapd;

import java.util.Enumeration;
import java.util.NoSuchElementException;

import com.instant802.cldc.channelplanner.WirelessNeighborhoodRecord;
import com.instant802.cldc.net.SocketException;

/**
 * @author robin
 */
public class HostAPDRadio {

	private static final String ITERATION_START = "start";
	private static final String ITERATION_FINISH = "finish";
	private static final String IFACE_BASE = "hostapd";
	private static final String CMD_AP_FIRST = "AP-FIRST";
	private static final String CMD_AP_NEXT = "AP-NEXT-";	
	private static final String FAIL = "FAIL";
	
	private final HostAPD hostAPD;
	private final String ifaceName;
	
	HostAPDRadio( HostAPD hostAPD, String ifaceName ) {
		this.hostAPD = hostAPD;
		this.ifaceName = ifaceName;
	}
	
	/**
	 * return an enumeration which allows the currently associated stations to be 
	 * traversed.
	 */
//	public Enumeration Stations() {		
//	}
	
	/**
	 * return an enumeration which allows the currently visible neighboring APs to be
	 * traversed
	 */
	public Enumeration Neighbors() {
		return new NeighborEnumeration();
	}
	
	private class NeighborEnumeration implements Enumeration {

		String lastReply = ITERATION_START;
		
		boolean findFirstElement() {
			// This is the first call to one of these routines.
			String apString;
			try {
				apString = APFirst();
			} catch (SocketException e) {
				return false;
			}
			if( isTerminator( apString ) ) {
				return false;
			}
			lastReply = apString;
			return true;			
		}
		
		void findNextElement( NeighboringAP ap ) {
			try {
				lastReply = APNext( ap.macAddress() );
				if( isTerminator( lastReply ) ) {
					lastReply = ITERATION_FINISH;
				}
			} catch (SocketException e) {
				lastReply = ITERATION_FINISH;
			}
			
		}
		
		public boolean hasMoreElements() {
			if( ITERATION_START.equals( lastReply ) ) {
				return findFirstElement();
			} else {
				// This is a subsequent call to one of these routines.
				return ! ITERATION_FINISH.equals( lastReply );
			}
		}

		/**
		 * return true if a particular reply should be treated as a terminator.
		 */
		boolean isTerminator( String reply ) {
			if( reply == null ) {
				return true;
			}
			if( reply == FAIL ) {
				return true;
			}
			if( reply.length() <= 1 ) {
				return true;
			}
			return false;
		}
		
		public Object nextElement() {
			NeighboringAP found;
			if( ITERATION_START.equals( lastReply ) ) {
				if( ! findFirstElement() ) {
					throw new NoSuchElementException( "the neighbor table is empty");
				}				
			}
			found = new WirelessNeighborhoodRecord( lastReply );
			// now we have the NeighboringAP that is found we need to get the next one and see if it is.
			findNextElement( found );
			return found;
		}
	}
	
	String APFirst() throws SocketException {
		return command( CMD_AP_FIRST );
	}
	
	String APNext( String mac ) throws SocketException {
		return command( CMD_AP_NEXT + mac );
	}
	
	private String command( String command ) throws SocketException {
		return hostAPD.requestResponse( getAbstractSocketName(), command );
	}
	
	/**
	 * return the string name for the abstract socket that
	 * handles commands relating to this interface 
	 */
	private String getAbstractSocketName() {
		return IFACE_BASE + "-" + ifaceName;
	}

	public String name() {
		return ifaceName;
	}
}
